class SliderPanel extends Element {
    var range, width, height, spacing, methods, spectra;
    var distribution, all_candidates, need_update;
    var selected_method;
    var candidates, switch_points, segments, last_ranks;

    function SliderPanel(
        parent, range, width, height, spacing, methods, distribution) {
        init(parent);
        this.range = range;
        this.width = width;
        this.height = height;
        this.spacing = spacing;
        this.methods = methods;
        this.distribution = distribution;
        this.selected_method = 0;

        spectra = [];
        for (var m = 0, y = 0; m < methods.length; m++, y += height + spacing) {
            spectra.push(new Spectrum(this, methods[m].name, y, width, height));
        }
        
        all_candidates = [];
        need_update = true;
        var slider_panel = this;
        clip.onEnterFrame = function() { slider_panel.update(); }
    }

    function set_distribution(distribution) {
        this.distribution = distribution;
        need_update = true;
    }

    function add_candidate(name, position, color, active) {
        var length = (height + spacing)*methods.length;
        var candidate = new Candidate(this, name, position, color, length, 12);
        candidate.clip._x = screen(position);
        candidate.set_active(active);
        all_candidates.push(candidate);
        need_update = true;
    }

    function world(x) {
        return (x*2/width - 1)*range;
    }
    
    function screen(position) {
        return ((position/range) + 1)*width/2;
    }
    
    function update() {
        for (var i = all_candidates.length - 1; i >= 0; i--) {
            if (all_candidates[i].pending_update) {
                all_candidates[i].pending_update = false;
                move(all_candidates[i], all_candidates[i].clip._x);
            }
        }
        if (need_update) {
            elect();
            need_update = false;
        }
        if (parent.voter_panel.pending_update) {
            update_example();
            parent.voter_panel.pending_update = false;
        }
    }
    
    function move(candidate, x) {
        candidate.position = world(x);
        need_update = true;
    }
    
    function elect() {
        candidates = [];
        for (var i = 0; i < all_candidates.length; i++) {
            if (all_candidates[i].active) {
                candidates.push(all_candidates[i]);
            } else {
                parent.blend_panel.set_blend(all_candidates[i], 0, 0);
            }
        }
        candidates.sortOn('position', Array.NUMERIC);
        
        switch_points = [];
        for (var i = 0; i < candidates.length - 1; i++) {
            for (var j = i + 1; j < candidates.length; j++) {
                switch_points.push({
                    pos: (candidates[i].position + candidates[j].position)/2,
                    from: i,
                    to: j
                });
            }
        }
        switch_points.sortOn('pos', Array.NUMERIC);

        var approval_blend = (methods[selected_method].name === 'Approval');
        if (approval_blend) {
            for (var c = 0; c < candidates.length; c++) {
                parent.blend_panel.set_blend(candidates[c],
                    screen(candidates[c].position - 1),
                    screen(candidates[c].position + 1));
            }
        }

        segments = [];
        var ranks = [0, 1, 2, 3, 4];
        var blend_start = world(0), favorite;
        ranks.length = candidates.length;
        for (var i = 0; i < switch_points.length; i++) {
            var point = switch_points[i];
            segments.push([point.pos, ranks.slice()]);
            if (ranks[point.from] === 0 && !approval_blend) {
                parent.blend_panel.set_blend(
                    candidates[point.from],
                    screen(blend_start), screen(point.pos));
                blend_start = point.pos;
            }
            ranks[point.from]++;
            ranks[point.to]--;
            if (ranks[point.to] === 0) {
                favorite = point.to;
            }
        }
        if (!approval_blend) {
            parent.blend_panel.set_blend(
                candidates[favorite], screen(blend_start), width);
        }
        last_ranks = ranks;
        
        // An optimization: assume that the candidate at an extreme end wins
        // for all points beyond that extreme.
        var n = candidates.length;
        var lower = Math.floor(
            screen((candidates[0].position + candidates[1].position)/2));
        var upper = Math.ceil(
            screen((candidates[n-2].position + candidates[n-1].position)/2));
        for (var m = 0; m < methods.length; m++) {
            if (spectra[m].active && methods[m].name === 'Borda') {
                lower = 0;
                upper = width;
            }
        }

        // Fill in the spectrum.
        for (var m = 0; m < methods.length; m++) {
            if (spectra[m].active) {
                spectra[m].setlower(lower, candidates[0].color);
                spectra[m].setupper(width - upper, candidates[n-1].color);
            } else {
                spectra[m].setlower(width, null);
                spectra[m].setupper(width, null);
            }
        }
        
        for (var x = lower - (lower % 2); x < upper; x+=2) {
            var centroid = world(x);
            var votes = collect_votes(centroid, segments, last_ranks);

            for (var m = 0; m < methods.length; m++) {
                if (spectra[m].active) {
                    var winner = methods[m].elect(votes, candidates, centroid);
                    spectra[m].set(x, winner.color);
                    spectra[m].set(x + 1, winner.color);
                }
            }
        }

        update_example();
    }

    function update_example() {
        // Color the dot.
        var centroid = world(parent.voter_panel.get_x());
        var votes = collect_votes(centroid, segments, last_ranks);
        var m = selected_method;
        var data = {};
        var winner = methods[m].elect(votes, candidates, centroid, data);
        votes = collect_votes(centroid, segments, last_ranks);
        parent.voter_panel.show_election(winner.color);
        parent.explanation_panel.show_election(
            methods[m], votes, candidates, centroid, data, distribution);
    }

    function collect_votes(centroid, segments, last_ranks) {
        var votes = [];
        var last_cum = 0, next_cum;

        for (var i = 0; i < segments.length; i++) {
            next_cum = distribution.cumulative(segments[i][0] - centroid);
            if (next_cum > last_cum) {
                votes.push({
                    count: next_cum - last_cum,
                    ranks: segments[i][1].slice()
                });
                last_cum = next_cum;
            }
        }
        
        votes.push({
            count: 1 - last_cum,
            ranks: last_ranks.slice()
        });
        return votes;
    }

    function select_point(spectrum, x) {
        var row = 0;
        for (var m = 0; m < methods.length; m++) {
            if (spectra[m] == spectrum) {
                row = m;
                selected_method = m;
            }
        }
        parent.voter_panel.move_to(x, row);
        need_update = true;
    }
}
